"""
Functions to get the network dict based on the config
via the API get_model, train_step, forward_step.

https://github.com/rwth-i6/returnn/issues/1120
"""

from __future__ import annotations
from typing import Any, Dict
from returnn.util.basic import BehaviorVersion
from returnn.tensor import TensorDict
from .. import frontend_layers as rfl


def get_net_dict(*, epoch: int, step: int, task: str) -> Dict[str, Any]:
    """called from the RETURNN config"""
    from returnn.config import get_global_config

    BehaviorVersion.set_min_behavior_version(rfl.min_returnn_behavior_version)

    config = get_global_config()
    rfl.Layer.reset_default_root()

    extern_data_dict = config.typed_value("extern_data")
    extern_data = TensorDict()
    extern_data.update(extern_data_dict, auto_convert=True)
    for data in extern_data.data.values():
        rfl.register_extern_data(data)

    get_model_func = config.typed_value("get_model")
    model = get_model_func(epoch=epoch, step=step)

    if task == "train":
        train_step_func = config.typed_value("train_step")
        train_step_func(
            model=model,
            extern_data=extern_data,
        )
    elif task == "forward":
        forward_step_func = config.typed_value("forward_step")
        forward_step_func(
            model=model,
            extern_data=extern_data,
        )
    else:
        raise ValueError(f"invalid task {task!r}")

    net_dict = rfl.Layer.top().root.get_returnn_config().get_net_dict_raw_dict(root_module=model)
    return net_dict



"""
Dict of tensors.

For example the extern_data in the user config is such a dict,
describing the data coming from the dataset (after batch preparation).

We also might have model_outputs in the user config.
(https://github.com/rwth-i6/returnn/issues/1166)
"""

from __future__ import annotations
from typing import Optional, Union, Any, Dict, Sequence
from .tensor import Tensor


_TensorT = Union[Tensor, Dict[str, Any]]
_DataAutoConvertT = Union[Dict[str, _TensorT], Sequence[_TensorT]]
_DataStrictT = Union[Dict[str, Tensor], Sequence[Tensor]]


class TensorDict:
    """dict of tensors"""

    def __init__(self, data: Optional[_DataStrictT] = None):
        self.data = {}  # type: Dict[str, Tensor]
        if data:
            self.update(data)

    def update(self, data: _DataAutoConvertT, *, auto_convert: bool = False):
        """update"""
        if isinstance(data, dict):
            for key, value in data.items():
                if auto_convert:
                    value = _convert_to_tensor(value, name=key)
                else:
                    assert isinstance(value, Tensor)
                self.data[key] = value
        elif isinstance(data, (list, tuple)):
            for value in data:
                if auto_convert:
                    value = _convert_to_tensor(value)
                else:
                    assert isinstance(value, Tensor)
                self.data[value.name] = value
        else:
            raise TypeError(f"invalid `data` type: {type(data)}")


def _convert_to_tensor(opts: _TensorT, *, name: Optional[str] = None) -> Tensor:
    """
    :param opts:
    """
    if isinstance(opts, Tensor):
        return opts
    assert isinstance(opts, dict)
    opts = opts.copy()
    if name:
        opts["name"] = name
    else:
        assert "name" in opts, f"missing `name` in Tensor {opts!r}"
    return Tensor(**opts)
